from pathlib import Path
from typing import TYPE_CHECKING, Callable
from uuid import uuid4

import yappi

if TYPE_CHECKING:
    from fastapi import FastAPI, Request, Response


def set_yappi_profiler(app: "FastAPI") -> None:
    @app.middleware("http")
    async def profile_process_time(
        request: "Request",
        call_next: Callable[["Request"], "Response"],
    ) -> "Response":
        yappi.set_clock_type("cpu")
        yappi.start()

        response = await call_next(request)

        yappi.stop()

        with Path(__file__).parents[2] / "profiling" as path:
            folder_location = path
            folder_location.mkdir(parents=True, exist_ok=True)

        profiling_log_id = uuid4()

        with Path(f"{folder_location}/{profiling_log_id}.log").open(mode="w") as file:
            stats = yappi.get_func_stats()
            stats.print_all(
                out=file,
                columns={
                    0: ("name", 180),
                    1: ("ncall", 8),
                    2: ("tsub", 8),
                    3: ("ttot", 8),
                    4: ("tavg", 8),
                },
            )

            response.headers["profiling-log-id"] = str(profiling_log_id)

        return response



from fastapi import APIRouter, Depends
from pydantic import UUID4
from sqlalchemy.orm import Session

from yak_server.config_file import Settings, get_settings
from yak_server.database.models import UserModel
from yak_server.helpers.rules import RULE_MAPPING
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.database import get_db
from yak_server.v1.helpers.errors import RuleNotFound
from yak_server.v1.models.generic import GenericOut

router = APIRouter(prefix="/rules", tags=["rules"])


@router.post("/{rule_id}")
def execute_rule(
    rule_id: UUID4,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[None]:
    found_rule = None

    for rule in settings.rules.__root__:
        if rule.id == rule_id:
            found_rule = rule
            break

    if found_rule is None:
        raise RuleNotFound(rule_id)

    rule_config = found_rule.config
    rule_function = RULE_MAPPING[found_rule.id]

    rule_function(db, user, rule_config)

    return GenericOut(result=None)



"""Change description to description_fr

Revision ID: 6e28ca2ef0c7
Revises: 314114c0b886
Create Date: 2023-05-30 20:57:17.927987

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "6e28ca2ef0c7"
down_revision = "314114c0b886"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("group", sa.Column("description_fr", sa.String(length=100), nullable=False))
    op.drop_index("description", table_name="group")
    op.create_unique_constraint(None, "group", ["description_fr"])
    op.drop_column("group", "description")
    op.add_column("phase", sa.Column("description_fr", sa.String(length=100), nullable=False))
    op.drop_column("phase", "description")
    op.add_column("team", sa.Column("description_fr", sa.String(length=100), nullable=False))
    op.drop_index("description", table_name="team")
    op.create_unique_constraint(None, "team", ["description_fr"])
    op.drop_column("team", "description")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("team", sa.Column("description", mysql.VARCHAR(length=100), nullable=False))
    op.drop_constraint(None, "team", type_="unique")
    op.create_index("description", "team", ["description"], unique=False)
    op.drop_column("team", "description_fr")
    op.add_column("phase", sa.Column("description", mysql.VARCHAR(length=100), nullable=False))
    op.drop_column("phase", "description_fr")
    op.add_column("group", sa.Column("description", mysql.VARCHAR(length=100), nullable=False))
    op.drop_constraint(None, "group", type_="unique")
    op.create_index("description", "group", ["description"], unique=False)
    op.drop_column("group", "description_fr")
    # ### end Alembic commands ###



from itertools import chain
from typing import TYPE_CHECKING
from uuid import UUID

from sqlalchemy import and_

from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    MatchModel,
    PhaseModel,
)

from .group_position import get_group_rank_with_code

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from yak_server.database.models import UserModel


def compute_finale_phase_from_group_rank(
    db: "Session",
    user: "UserModel",
    rule_config: dict,
) -> None:
    first_phase_phase_group = db.query(GroupModel).filter_by(code=rule_config["to_group"]).first()

    groups_result = {
        group.code: get_group_rank_with_code(db, user, group.id)
        for group in db.query(GroupModel)
        .join(GroupModel.phase)
        .filter(
            PhaseModel.code == rule_config["from_phase"],
        )
    }

    for index, match_config in enumerate(rule_config["versus"], 1):
        if all(
            team.played == len(groups_result[match_config["team1"]["group"]]) - 1
            for team in chain(
                groups_result[match_config["team1"]["group"]],
                groups_result[match_config["team2"]["group"]],
            )
        ):
            team1 = groups_result[match_config["team1"]["group"]][
                match_config["team1"]["rank"] - 1
            ].team
            team2 = groups_result[match_config["team2"]["group"]][
                match_config["team2"]["rank"] - 1
            ].team

            binary_bet = (
                db.query(BinaryBetModel)
                .join(BinaryBetModel.match)
                .filter(
                    and_(
                        MatchModel.index == index,
                        BinaryBetModel.user_id == user.id,
                        MatchModel.group_id == first_phase_phase_group.id,
                    ),
                )
                .first()
            )

            binary_bet.match.team1_id = team1.id
            binary_bet.match.team2_id = team2.id

            db.flush()

    db.commit()


RULE_MAPPING = {
    UUID("492345de-8d4a-45b6-8b94-d219f2b0c3e9"): compute_finale_phase_from_group_rank,
}



"""Add internal flag field in team table

Revision ID: 045a9142794e
Revises: 9b91a1a350a0
Create Date: 2023-04-18 23:28:37.358017

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "045a9142794e"
down_revision = "9b91a1a350a0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("team", schema=None) as batch_op:
        batch_op.add_column(sa.Column("internal_flag_url", sa.String(length=300), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("team", schema=None) as batch_op:
        batch_op.drop_column("internal_flag_url")

    # ### end Alembic commands ###



from fastapi import APIRouter, Depends
from sqlalchemy.orm import Session

from yak_server.database.models import GroupModel, PhaseModel, UserModel
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.database import get_db
from yak_server.v1.helpers.errors import GroupNotFound, PhaseNotFound
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.groups import (
    AllGroupsResponse,
    GroupOut,
    GroupResponse,
    GroupsByPhaseCodeResponse,
    GroupWithPhaseIdOut,
)
from yak_server.v1.models.phases import PhaseOut

router = APIRouter(prefix="/groups", tags=["groups"])


@router.get("/")
def retrieve_all_groups(
    lang: Lang = DEFAULT_LANGUAGE,
    _: UserModel = Depends(get_current_user),
    db: Session = Depends(get_db),
) -> GenericOut[AllGroupsResponse]:
    groups = db.query(GroupModel).order_by(GroupModel.index)
    phases = db.query(PhaseModel).order_by(PhaseModel.index)

    return GenericOut(
        result=AllGroupsResponse(
            phases=[PhaseOut.from_instance(phase, lang) for phase in phases],
            groups=[GroupWithPhaseIdOut.from_instance(group, lang) for group in groups],
        ),
    )


@router.get("/{group_code}")
def retrieve_group_by_id(
    group_code: str,
    lang: Lang = DEFAULT_LANGUAGE,
    _: UserModel = Depends(get_current_user),
    db: Session = Depends(get_db),
) -> GenericOut[GroupResponse]:
    group = db.query(GroupModel).filter_by(code=group_code).first()

    if not group:
        raise GroupNotFound(group_code)

    return GenericOut(
        result=GroupResponse(
            phase=PhaseOut.from_instance(group.phase, lang),
            group=GroupOut.from_instance(group, lang),
        ),
    )


@router.get("/phases/{phase_code}")
def retrieve_groups_by_phase_code(
    phase_code: str,
    lang: Lang = DEFAULT_LANGUAGE,
    _: UserModel = Depends(get_current_user),
    db: Session = Depends(get_db),
) -> GenericOut[GroupsByPhaseCodeResponse]:
    phase = db.query(PhaseModel).filter_by(code=phase_code).first()

    if not phase:
        raise PhaseNotFound(phase_code)

    groups = db.query(GroupModel).order_by(GroupModel.index).filter_by(phase_id=phase.id)

    return GenericOut(
        result=GroupsByPhaseCodeResponse(
            phase=PhaseOut.from_instance(phase, lang),
            groups=[GroupOut.from_instance(group, lang) for group in groups],
        ),
    )



from pydantic import UUID4, BaseModel, Extra


class SignupIn(BaseModel, extra=Extra.forbid):
    name: str
    first_name: str
    last_name: str
    password: str


class SignupOut(BaseModel, extra=Extra.forbid):
    id: UUID4
    name: str
    token: str


class LoginIn(BaseModel, extra=Extra.forbid):
    name: str
    password: str


class LoginOut(BaseModel, extra=Extra.forbid):
    id: UUID4
    name: str
    token: str


class CurrentUserOut(BaseModel):
    id: UUID4
    name: str

    class Config:
        orm_mode = True


class ModifyUserIn(BaseModel, extra=Extra.forbid):
    password: str



from fastapi import APIRouter, Depends
from fastapi.responses import RedirectResponse
from pydantic import UUID4
from sqlalchemy.orm import Session

from yak_server.database.models import TeamModel
from yak_server.helpers.format import is_iso_3166_1_alpha_2_code, is_uuid4
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang
from yak_server.v1.helpers.database import get_db
from yak_server.v1.helpers.errors import InvalidTeamId, TeamNotFound
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.teams import AllTeamsResponse, OneTeamResponse, TeamOut

router = APIRouter(prefix="/teams", tags=["teams"])


@router.get("/")
def retrieve_all_teams(
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
) -> GenericOut[AllTeamsResponse]:
    return GenericOut(
        result=AllTeamsResponse(
            teams=[TeamOut.from_instance(team, lang) for team in db.query(TeamModel).all()],
        ),
    )


@router.get("/{team_id}")
def retrieve_team_by_id(
    team_id: str,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
) -> GenericOut[OneTeamResponse]:
    if is_uuid4(team_id):
        team = db.query(TeamModel).filter_by(id=team_id).first()
    elif is_iso_3166_1_alpha_2_code(team_id):
        team = db.query(TeamModel).filter_by(code=team_id).first()
    else:
        raise InvalidTeamId(team_id)

    if not team:
        raise TeamNotFound(team_id)

    return GenericOut(result=OneTeamResponse(team=TeamOut.from_instance(team, lang)))


@router.get("/{team_id}/flag")
def retrieve_team_flag_by_id(team_id: UUID4, db: Session = Depends(get_db)) -> RedirectResponse:
    team = db.query(TeamModel).filter_by(id=str(team_id)).first()

    if not team:
        raise TeamNotFound(team_id)

    return RedirectResponse(team.internal_flag_url)



from itertools import chain
from typing import List

from fastapi import APIRouter, Depends
from sqlalchemy import and_
from sqlalchemy.orm import Session

from yak_server.config_file import Settings, get_settings
from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    UserModel,
)
from yak_server.helpers.group_position import get_group_rank_with_code
from yak_server.v1.helpers.auth import get_admin_user, get_current_user
from yak_server.v1.helpers.database import get_db
from yak_server.v1.helpers.errors import NoResultsForAdminUser
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.results import UserResult

router = APIRouter(
    tags=["results"],
)


@router.get("/score_board")
def retrieve_score_board(
    _: UserModel = Depends(get_current_user),
    db: Session = Depends(get_db),
) -> GenericOut[List[UserResult]]:
    return GenericOut(
        result=[
            UserResult.from_instance(user, rank)
            for rank, user in enumerate(
                db.query(UserModel)
                .order_by(UserModel.points.desc())
                .where(UserModel.name != "admin"),
                1,
            )
        ],
    )


@router.get("/results")
def retrieve_user_results(
    user: UserModel = Depends(get_current_user),
    db: Session = Depends(get_db),
) -> GenericOut[UserResult]:
    if user.name == "admin":
        raise NoResultsForAdminUser

    user_result = UserResult.from_instance(
        user,
        [
            index
            for index, user_result in enumerate(
                db.query(UserModel)
                .order_by(UserModel.points.desc())
                .filter(
                    UserModel.name != "admin",
                ),
                1,
            )
            if user_result.id == user.id
        ][0],
    )

    return GenericOut(result=user_result)


@router.post("/compute_points")
def compute_pointsby_by_admin(
    _: UserModel = Depends(get_admin_user),
    db: Session = Depends(get_db),
    settings: Settings = Depends(get_settings),
) -> GenericOut[List[UserResult]]:
    users = compute_points(
        db,
        settings.base_correct_result,
        settings.multiplying_factor_correct_result,
        settings.base_correct_score,
        settings.multiplying_factor_correct_score,
        settings.team_qualified,
        settings.first_team_qualified,
    )

    return GenericOut(
        result=[UserResult.from_instance(user, rank) for rank, user in enumerate(users, 1)],
    )


def compute_points(
    db: Session,
    base_correct_result: int,
    multiplying_factor_correct_result: int,
    base_correct_score: int,
    multiplying_factor_correct_score: int,
    team_qualified: int,
    first_team_qualified: int,
) -> None:
    admin = db.query(UserModel).filter_by(name="admin").first()

    results = []

    for real_score in admin.score_bets:
        number_correct_result = 0
        number_correct_score = 0
        user_ids_found_correct_result = []
        user_ids_found_correct_score = []

        for user_score in (
            db.query(ScoreBetModel)
            .join(ScoreBetModel.match)
            .filter(
                and_(
                    ScoreBetModel.user_id != admin.id,
                    MatchModel.index == real_score.match.index,
                    MatchModel.group_id == real_score.match.group_id,
                ),
            )
        ):
            if user_score.is_same_results(real_score):
                number_correct_result += 1
                user_ids_found_correct_result.append(user_score.user_id)

                if user_score.is_same_scores(real_score):
                    number_correct_score += 1
                    user_ids_found_correct_score.append(user_score.user_id)

        results.append(
            {
                "number_correct_result": number_correct_result,
                "user_ids_found_correct_result": user_ids_found_correct_result,
                "number_correct_score": number_correct_score,
                "user_ids_found_correct_score": user_ids_found_correct_score,
            },
        )

    result_groups = {}

    users = db.query(UserModel).filter(UserModel.name != "admin")

    for group in (
        db.query(GroupModel)
        .join(GroupModel.phase)
        .filter(
            PhaseModel.code == "GROUP",
        )
    ):
        group_result_admin = get_group_rank_with_code(db, admin, group.id)

        if all_results_filled_in_group(group_result_admin):
            admin_first_team_id = group_result_admin[0].team.id
            admin_second_team_id = group_result_admin[1].team.id

            for user in users:
                if user.id not in result_groups:
                    result_groups[user.id] = {
                        "number_qualified_teams_guess": 0,
                        "number_first_qualified_guess": 0,
                    }

                group_result_user = get_group_rank_with_code(db, user, group.id)

                if all_results_filled_in_group(group_result_user):
                    user_first_team_id = group_result_user[0].team.id
                    user_second_team_id = group_result_user[1].team.id

                    result_groups[user.id]["number_qualified_teams_guess"] += len(
                        {user_first_team_id, user_second_team_id}
                        & {admin_first_team_id, admin_second_team_id},
                    )

                    if user_first_team_id == admin_first_team_id:
                        result_groups[user.id]["number_first_qualified_guess"] += 1

    quarter_finals_team = team_from_group_code(admin, "4")
    semi_finals_team = team_from_group_code(admin, "2")
    final_team = team_from_group_code(admin, "1")
    winner = winner_from_user(admin)

    numbers_of_players = users.count()

    for user in users:
        user.number_score_guess = 0
        user.number_match_guess = 0
        user.points = 0

        for result in results:
            if user.id in result["user_ids_found_correct_result"]:
                user.number_match_guess += 1
                user.points += base_correct_result + multiplying_factor_correct_result * (
                    numbers_of_players - result["number_correct_result"]
                ) / (numbers_of_players - 1)

            if user.id in result["user_ids_found_correct_score"]:
                user.number_score_guess += 1
                user.points += base_correct_score + multiplying_factor_correct_score * (
                    numbers_of_players - result["number_correct_score"]
                ) / (numbers_of_players - 1)

        user.number_qualified_teams_guess = result_groups.get(user.id, {}).get(
            "number_qualified_teams_guess",
            0,
        )
        user.number_first_qualified_guess = result_groups.get(user.id, {}).get(
            "number_first_qualified_guess",
            0,
        )

        user.points += user.number_qualified_teams_guess * team_qualified
        user.points += user.number_first_qualified_guess * first_team_qualified

        user.number_quarter_final_guess = len(
            team_from_group_code(user, "4").intersection(quarter_finals_team),
        )
        user.number_semi_final_guess = len(
            team_from_group_code(user, "2").intersection(semi_finals_team),
        )
        user.number_final_guess = len(
            team_from_group_code(user, "1").intersection(final_team),
        )
        user.number_winner_guess = len(winner_from_user(user).intersection(winner))

        user.points += 30 * user.number_quarter_final_guess
        user.points += 60 * user.number_semi_final_guess
        user.points += 120 * user.number_final_guess
        user.points += 200 * user.number_winner_guess

    db.commit()

    return users


def all_results_filled_in_group(group_result: list) -> bool:
    return all(team.played == len(group_result) - 1 for team in group_result)


def team_from_group_code(user: UserModel, group_code: str) -> set:
    return set(
        chain(
            *(
                (bet.match.team1.id, bet.match.team2.id)
                for bet in user.binary_bets.filter(GroupModel.code == group_code)
                .join(BinaryBetModel.match)
                .join(MatchModel.group)
                if bet.match.team1 is not None and bet.match.team2 is not None
            ),
        ),
    )


def winner_from_user(user: UserModel) -> set:
    finale_bet = list(
        user.binary_bets.filter(GroupModel.code == "1")
        .join(BinaryBetModel.match)
        .join(MatchModel.group),
    )

    if not finale_bet:
        return set()

    finale_bet = finale_bet[0]

    if finale_bet.is_one_won is None:
        return set()

    return {
        finale_bet.match.team1.id if finale_bet.is_one_won else finale_bet.match.team2.id,
    }



"""Remove locked column from score bet and binary bet.

Revision ID: c2ecfb568236
Revises: d75e76959af8
Create Date: 2023-03-12 06:30:53.548293

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "c2ecfb568236"
down_revision = "d75e76959af8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("binary_bet", schema=None) as batch_op:
        batch_op.drop_column("locked")

    with op.batch_alter_table("score_bet", schema=None) as batch_op:
        batch_op.drop_column("locked")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("score_bet", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("locked", mysql.TINYINT(display_width=1), autoincrement=False, nullable=True),
        )

    with op.batch_alter_table("binary_bet", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("locked", mysql.TINYINT(display_width=1), autoincrement=False, nullable=True),
        )

    # ### end Alembic commands ###



from typing import Generic, TypeVar

from pydantic import ConstrainedInt, Extra
from pydantic.generics import GenericModel

Result = TypeVar("Result")


class GenericOut(GenericModel, Generic[Result], extra=Extra.forbid):
    ok: bool = True
    result: Result


class PositiveOrZeroInt(ConstrainedInt):
    gt = -1



"""Add bet type from match.

Revision ID: b36763d4cb42
Revises: c2ecfb568236
Create Date: 2023-03-16 21:34:07.705206

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b36763d4cb42"
down_revision = "c2ecfb568236"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("match", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "bet_type_from_match",
                sa.Enum("SCORE_BET", "BINARY_BET", name="betmapping"),
                nullable=True,
            ),
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("match", schema=None) as batch_op:
        batch_op.drop_column("bet_type_from_match")

    # ### end Alembic commands ###



import logging

from fastapi import APIRouter, Depends
from pydantic import UUID4
from sqlalchemy import update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

from yak_server.config_file import Settings, get_settings
from yak_server.database.models import (
    GroupPositionModel,
    MatchModel,
    ScoreBetModel,
    UserModel,
)
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang, get_language_description
from yak_server.helpers.logging import modify_score_bet_successfully
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.database import get_db
from yak_server.v1.helpers.errors import (
    BetNotFound,
    GroupNotFound,
    LockedScoreBet,
    TeamNotFound,
)
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.groups import GroupOut
from yak_server.v1.models.phases import PhaseOut
from yak_server.v1.models.score_bets import (
    ModifyScoreBetIn,
    ScoreBetIn,
    ScoreBetOut,
    ScoreBetResponse,
)
from yak_server.v1.models.teams import FlagOut, TeamWithScoreOut

logger = logging.getLogger(__name__)

router = APIRouter(
    prefix="/score_bets",
    tags=["score_bets"],
)


def send_response(
    score_bet: ScoreBetModel,
    locked: bool,
    lang: Lang,
) -> GenericOut[ScoreBetResponse]:
    return GenericOut(
        result=ScoreBetResponse(
            phase=PhaseOut.from_instance(score_bet.match.group.phase, lang),
            group=GroupOut.from_instance(score_bet.match.group, lang),
            score_bet=ScoreBetOut(
                id=score_bet.id,
                locked=locked,
                team1=TeamWithScoreOut(
                    id=score_bet.match.team1.id,
                    code=score_bet.match.team1.code,
                    description=get_language_description(score_bet.match.team1, lang),
                    flag=FlagOut(url=score_bet.match.team1.flag_url),
                    score=score_bet.score1,
                ),
                team2=TeamWithScoreOut(
                    id=score_bet.match.team2.id,
                    code=score_bet.match.team2.code,
                    description=get_language_description(score_bet.match.team2, lang),
                    flag=FlagOut(url=score_bet.match.team2.flag_url),
                    score=score_bet.score2,
                ),
            ),
        ),
    )


@router.post("/")
def create_score_bet(
    score_bet_in: ScoreBetIn,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[ScoreBetResponse]:
    if is_locked(user.name, settings.lock_datetime):
        raise LockedScoreBet

    match = MatchModel(
        team1_id=score_bet_in.team1.id,
        team2_id=score_bet_in.team2.id,
        index=score_bet_in.index,
        group_id=score_bet_in.group.id,
    )

    db.add(match)
    try:
        db.flush()
    except IntegrityError as integrity_error:
        if "FOREIGN KEY (`team1_id`)" in str(integrity_error):
            raise TeamNotFound(team_id=score_bet_in.team1.id) from integrity_error

        if "FOREIGN KEY (`team2_id`)" in str(integrity_error):
            raise TeamNotFound(team_id=score_bet_in.team2.id) from integrity_error

        raise GroupNotFound(group_id=score_bet_in.group.id) from integrity_error

    score_bet = ScoreBetModel(
        match_id=match.id,
        user_id=user.id,
        score1=score_bet_in.team1.score,
        score2=score_bet_in.team2.score,
    )

    db.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet_in.team1.id,
            GroupPositionModel.user_id == user.id,
        ),
    )
    db.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet_in.team2.id,
            GroupPositionModel.user_id == user.id,
        ),
    )

    db.add(score_bet)
    db.commit()
    db.refresh(score_bet)

    return send_response(score_bet, is_locked(user.name, settings.lock_datetime), lang)


@router.get("/{bet_id}")
def retrieve_score_bet_by_id(
    bet_id: UUID4,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[ScoreBetResponse]:
    score_bet = db.query(ScoreBetModel).filter_by(user_id=user.id, id=str(bet_id)).first()

    if not score_bet:
        raise BetNotFound(bet_id)

    return send_response(score_bet, is_locked(user.name, settings.lock_datetime), lang)


@router.patch("/{bet_id}")
def modify_score_bet(
    bet_id: UUID4,
    modify_score_bet_in: ModifyScoreBetIn,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[ScoreBetResponse]:
    if is_locked(user.name, settings.lock_datetime):
        raise LockedScoreBet

    score_bet = (
        db.query(ScoreBetModel).filter_by(user_id=user.id, id=str(bet_id)).with_for_update().first()
    )

    if not score_bet:
        raise BetNotFound(bet_id)

    if (
        score_bet.score1 == modify_score_bet_in.team1.score
        and score_bet.score2 == modify_score_bet_in.team2.score
    ):
        return send_response(score_bet, is_locked(user.name, settings.lock_datetime), lang)

    logger.info(
        modify_score_bet_successfully(
            user.name,
            score_bet,
            modify_score_bet_in.team1.score,
            modify_score_bet_in.team2.score,
        ),
    )

    db.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet.match.team1_id,
            GroupPositionModel.user_id == user.id,
        ),
    )
    db.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet.match.team2_id,
            GroupPositionModel.user_id == user.id,
        ),
    )

    score_bet.score1 = modify_score_bet_in.team1.score
    score_bet.score2 = modify_score_bet_in.team2.score
    db.commit()

    return send_response(score_bet, is_locked(user.name, settings.lock_datetime), lang)


@router.delete("/{bet_id}")
def delete_score_bet_by_id(
    bet_id: UUID4,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[ScoreBetResponse]:
    if is_locked(user.name, settings.lock_datetime):
        raise LockedScoreBet

    score_bet = db.query(ScoreBetModel).filter_by(user_id=user.id, id=str(bet_id)).first()

    if not score_bet:
        raise BetNotFound(bet_id)

    response = send_response(score_bet, is_locked(user.name, settings.lock_datetime), lang)

    db.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet.match.team1_id,
            GroupPositionModel.user_id == user.id,
        ),
    )
    db.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet.match.team2_id,
            GroupPositionModel.user_id == user.id,
        ),
    )

    db.delete(score_bet)
    db.commit()

    return response



from typing import List

from fastapi import APIRouter, Depends
from pydantic import UUID4
from sqlalchemy.orm import Session

from yak_server.database.models import PhaseModel, UserModel
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.database import get_db
from yak_server.v1.helpers.errors import PhaseNotFound
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.phases import PhaseOut

router = APIRouter(prefix="/phases", tags=["phases"])


@router.get("/")
def retrieve_all_phases(
    lang: Lang = DEFAULT_LANGUAGE,
    _: UserModel = Depends(get_current_user),
    db: Session = Depends(get_db),
) -> GenericOut[List[PhaseOut]]:
    return GenericOut(
        result=[
            PhaseOut.from_instance(phase, lang)
            for phase in db.query(PhaseModel).order_by(PhaseModel.index)
        ],
    )


@router.get("/{phase_id}")
def retrieve_phase(
    phase_id: UUID4,
    lang: Lang = DEFAULT_LANGUAGE,
    _: UserModel = Depends(get_current_user),
    db: Session = Depends(get_db),
) -> GenericOut[PhaseOut]:
    phase = db.query(PhaseModel).filter_by(id=str(phase_id)).first()

    if not phase:
        raise PhaseNotFound(phase_id)

    return GenericOut(
        result=PhaseOut.from_instance(phase, lang),
    )



"""Introduce match reference table and allow nullable team_id foreign key in match table.

Revision ID: 9b91a1a350a0
Revises: e0f551c7766c
Create Date: 2023-04-07 23:06:16.326481

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "9b91a1a350a0"
down_revision = "e0f551c7766c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "match_reference",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("group_id", sa.String(length=100), nullable=False),
        sa.Column("index", sa.Integer(), nullable=False),
        sa.Column("team1_id", sa.String(length=100), nullable=True),
        sa.Column("team2_id", sa.String(length=100), nullable=True),
        sa.Column(
            "bet_type_from_match",
            sa.Enum("SCORE_BET", "BINARY_BET", name="betmapping"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["group_id"],
            ["group.id"],
        ),
        sa.ForeignKeyConstraint(
            ["team1_id"],
            ["team.id"],
        ),
        sa.ForeignKeyConstraint(
            ["team2_id"],
            ["team.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("match", schema=None) as batch_op:
        batch_op.alter_column("team1_id", existing_type=mysql.VARCHAR(length=100), nullable=True)
        batch_op.alter_column("team2_id", existing_type=mysql.VARCHAR(length=100), nullable=True)
        batch_op.drop_column("bet_type_from_match")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("match", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("bet_type_from_match", mysql.ENUM("SCORE_BET", "BINARY_BET"), nullable=True),
        )
        batch_op.alter_column("team2_id", existing_type=mysql.VARCHAR(length=100), nullable=False)
        batch_op.alter_column("team1_id", existing_type=mysql.VARCHAR(length=100), nullable=False)

    op.drop_table("match_reference")
    # ### end Alembic commands ###
